#!/usr/local/bin/perl
# -*- perl -*-

# Copyright (c) 2002 by Jeff Weisberg
# Author: Jeff Weisberg <argus @ tcp4me.com>
# Date: 2002-Apr-03 18:59 (EST)
# Function: testing of SNMP things
#
# $Id: SNMP.pm,v 1.12 2002/10/31 18:13:10 jaw Exp $

package SNMP;
use BER;

@ISA = qw(UDP);

my $snmpid = rand( 1000000 );
sub get_request  { 0 | context_flag };

my %oids =
(
 AdminStatus   => '.1.3.6.1.2.1.2.2.1.7',
 OperStatus    => '.1.3.6.1.2.1.2.2.1.8',
 ifInErrors    => '1.3.6.1.2.1.2.2.1.14',
 ifOutErrors   => '1.3.6.1.2.1.2.2.1.20',
 ifInOctets    => '.1.3.6.1.2.1.2.2.1.10',
 ifOutOctets   => '.1.3.6.1.2.1.2.2.1.16',
 BGPPeerStatus => '.1.3.6.1.2.1.15.3.1.2',
 ciscoEnvMonTemperatureStatusValue => '.1.3.6.1.4.1.9.9.13.1.3.1.3',
 # ...
 
 );

$doc = {
    package => __PACKAGE__,
    file    => __FILE__,
    isa     => [qw(UDP Service MonEl BaseIO)],
    methods => {},
    fields => {
      snmp::community => {
	  descr => 'SNMP community',
	  attrs => ['config', 'inherit'],
      },
      snmp::snmpcalc => {
	  descr => 'manipulate value in some manner',
	  attrs => ['config'],
      },
      snmp::alpha => {
	  descr => 'time constant for decaying average calculation',
	  # ave = (alpha * ave + value) / (alpha + 1)
	  attrs => ['config', 'inherit'],
	  default => '1',
      },
      snmp::calcdata => {},
      snmp::currvalue => {
	  descr => 'most recent value returned by query',
      },
      snmp::oid => {
	  descr => 'SNMP OID to query',
	  attrs => ['config'],
      },
      # test value fields
      snmp::expect => {
	  descr => 'SNMP test value - fail test if value does not match regex',
	  attrs => ['config'],
      },
      snmp::minvalue => {
	  descr => 'SNMP test value - fail test if value falls below this',
	  attrs => ['config'],
      },
      snmp::maxvalue => {
	  descr => 'SNMP test value - fail test if value rises above this',
	  attrs => ['config'],
      },
      snmp::eqvalue => {
	  descr => 'SNMP test value - fail test if value does not equal this',
	  attrs => ['config'],
      },
      snmp::nevalue => {
	  descr => 'SNMP test value - fail test if value equals this',
	  attrs => ['config'],
      },
    },
};

sub probe {
    my $name = shift;

    return ( $name =~ /^UDP\/SNMP/ ) ? [ 8, \&config ] : undef;
}

sub config {
    my $me = shift;
    my $cf = shift;

    bless $me;
    $me->init_from_config( $doc, 'snmp' );
    
    if( $me->{name} =~ /SNMP\/(.*)/ ){
	$me->{snmp}{oid} ||= $1;
    }
    
    $me->{label_right_maybe} ||= $me->{snmp}{oid};
    $me->SUPER::config($cf);
    $me->{udp}{port} ||= 161;
    $me->{snmp}{alpha} = 1 unless $me->{snmp}{alpha} > 0;
    $me->snmp_init($cf);
    $me->{uname} = "SNMP_$me->{snmp}{oid}_$me->{srvc}{hostname}";
    $me;
}

sub snmp_init {
    my $me = shift;
    my $cf = shift;
    my( $req, $oid );

    $oid = $me->{snmp}{oid};

    return $cf->error( "invalid OID ($oid)" )
	unless $oid;

    $oid =~ s/([^\d\.]+)/$oids{$1}||'UNKNOWN'/ge;
    
    return $cf->error( "invalid OID ($oid)" )
	unless( $oid =~ /^[\d\.]+$/ );

    $me->{snmp}{oid} = $oid;
    
    # Why lifts she up her arms in sequence thus?
    #   -- Shakespeare, Titus Andronicus
    # SNMP v1 packet
    $req = encode_sequence( encode_int_0,
		    encode_string( $me->{snmp}{community} ), 
		    encode_tagged_sequence(
				   get_request,
				   encode_int( $snmpid++ ),
				   encode_int_0,
				   encode_int_0,
				   encode_sequence(
					   encode_sequence(
						   encode_oid(split /\./, $oid ),
						   encode_null()))) );
    $me->{udp}{send} = $req;
    $me;
}

sub readable {
    my $me = shift;
    my( $fh, $i, $l );
    my( $foo, $transient );

    $fh = $me->{fd};
    $i = recv($fh, $l, 8192, 0);
    return $me->isdown( "SNMP recv failed: $!" )
	unless defined($i);

    $me->debug( "SNMP recv data" );
    $me->{udp}{rbuffer} = $l;		# for debugging

    $BER::errmsg = '';
    
    my( $ver, $comm, $reqid, $errst, $errin, $oid, $value ) =
	decode_by_template($l, "%{%i%s%*{%i%i%i%{%{%O%@", 162 );

    $value = pretty_print($value);

    if( $BER::errmsg ){
	return $me->isdown( "BER error: $BER::errmsg" );
    }

    if( $value eq "(null)" ){
	return $me->isdown( "BER error: NULL (no such OID?)" );
    }
    
    $me->debug( "SNMP recv - raw value $value" );
    
    if( $me->{snmp}{snmpcalc} && $me->{snmp}{snmpcalc} =~ /rate/ ){
	if( $me->{snmp}{calcdata}{lastt} && $me->{snmp}{calcdata}{lastt} != $^T ){
	    if( $value < $me->{snmp}{calcdata}{lastv} ){
		# handle counter wrap-around
		$foo = 0xFFFFFFFF - $me->{snmp}{calcdata}{lastv};
		$foo += $value + 1;
	    }else{
		$foo = $value - $me->{snmp}{calcdata}{lastv};
	    }
	    $foo /= ($^T - $me->{snmp}{calcdata}{lastt});
	    $me->debug( "SNMP rate $foo" );
	}else{
	    $transient = 1;
	    $foo = 0;
	    $me->debug( "SNMP rate startup" );
	}
	$me->{snmp}{calcdata}{lastv} = $value;
	$me->{snmp}{calcdata}{lastt} = $^T;
	$value = $foo;
	return $me->done() if $transient;
    }

    # Why birds and beasts from quality and kind,
    # Why old men fool and children calculate,
    #   -- Shakespeare, Julius Ceasar
    if( $me->{snmp}{snmpcalc} && $me->{snmp}{snmpcalc} =~ /ave/ ){
	# moving average
	my $x = $me->{snmp}{alpha} * $me->{snmp}{calcdata}{ave};
	$x += $value;
	$x /= $me->{snmp}{alpha} + 1;
	$me->{snmp}{calcdata}{ave} = $x;
	$value = $x;
	$me->debug( "SNMP average $value" );
    }

    if( $me->{snmp}{snmpcalc} && $me->{snmp}{snmpcalc} =~ /bits/ ){
	# convert Bytes -> bits
	$value *= 8;
    }
    
    if( $me->{snmp}{snmpcalc} ){
	$value = sprintf "%.2f", $value;
    }
    
    $me->{snmp}{currvalue} = $value;
    $me->{srvc}{result}    = $value;
    $me->debug( "SNMP recv - curr $value" );

    # test value
    if( defined($me->{snmp}{expect}) ){
	my $e = $me->{snmp}{expect};
	# Thou sober-suited matron, all in black,
	# And learn me how to lose a winning match,
	#   -- Shakespeare, Romeo+Juliet
	return $me->isdown( "SNMP did not match expected regex" )
	    unless $value =~ /$e/;
    }
    # Lesser than Macbeth, and greater.
    #   -- Shakespeare, Macbeth
    if( defined($me->{snmp}{minvalue}) && $value < $me->{snmp}{minvalue} ){
	return $me->isdown( "SNMP less than min" );
    }
    # The greater scorns the lesser
    #   -- Shakespeare, Timon of Athens
    if( defined($me->{snmp}{maxvalue}) && $value > $me->{snmp}{maxvalue} ){
	return $me->isdown( "SNMP more than max" );
    }
    # Repugnant to command: unequal match'd
    #   -- Shakespeare, Hamlet
    if( defined($me->{snmp}{eqvalue}) && $value != $me->{snmp}{eqvalue} ){
	return $me->isdown( "SNMP not equal" );
    }
    # In equal scale weighing delight and dole
    #   -- Shakespeare, Hamlet
    if( defined($me->{snmp}{nevalue}) && $value == $me->{snmp}{nevalue} ){
	return $me->isdown( "SNMP equal" );
    }
    return $me->isup();
}

sub webpage_more {
    my $me = shift;
    my $fh = shift;
    my( $k, $v );
    
    $me->SUPER::webpage_more($fh);
    
    foreach $k (qw(oid)){
	$v = $me->{snmp}{$k};
	print $fh "<TR><TD>SNMP $k</TD><TD>$v</TD></TR>\n" if defined($v);
    }
}

sub about_more {
    my $me = shift;
    my $ctl = shift;
    my( $k, $v );

    $me->SUPER::about_more($ctl);
    
    foreach $k (sort keys %{$me->{snmp}}){
	$v = $me->{snmp}{$k};
	next unless defined $v;
	if( ref($v) ){
	    $v = "#<REF>";
	}else{
	    $v = ::xxx_encode( $v );
	}
	$ctl->write( "snmp::$k:\t$v\n" );
    }
}


Doc::register( $doc );
push @Service::probes, \&probe;

1;
